# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/15_loopstack.ipynb (unless otherwise specified).

__all__ = ['create_event', 'events', 'LoopStack', 'train_callbacks', 'train_single_forward', 'TrainLoop', 'EvalLoop']

# Cell
from .loop import Loop,ProgressBar,Tolerate,Event,Stuff
from types import MethodType
import numpy as np

# Cell
def create_event(event_name):
    class BatchEvent(Event):pass
    BatchEvent.__name__ = event_name
    return BatchEvent

def events(*enames):
    return list(map(create_event,enames))

# Cell
class LoopStack(Loop):
    settings = []
    """
    A stack of loop
    """
    @classmethod
    def from_loops(cls,*loops):
        def init(self,iterable=[],name = None):
            name = name if name!=None else cls.__name__

            self.loops = dict()
            l = Loop(iterable)
            for L in loops:
                l = L(iterable = l)
            super().__init__(iterable = l,
                             name = name)

            for stuff in cls.settings:
                self.make_stuff(stuff)

        setattr(cls,"__init__",init)
        return cls

    @classmethod
    def new_setting(cls,*settings):
        cls.settings+=list(settings)

    def make_stuff(self,name):
        new_stuff = Stuff(name)
        setattr(self.core,name,new_stuff)
        setattr(self,name,new_stuff)

    def __repr__(self,):
        return f"LoopStack>:{self.name}\n\t"+\
            "\n\t".join(map(str,self.core.layers[:-1]))

# Cell
import torch

def train_callbacks(loop):
    @loop.on_DATA_PROCESS
    def opt_zero_grad(loop):
        loop.opt("zero_grad")

    @loop.before_1st_FORWARD
    def switch_model_to_train(loop):
        loop.model("train")

    @loop.BACKWARD.on
    def opt_step(loop):
        loop.loss("backward")

    @loop.BACKWARD.on
    def opt_step(loop):
        loop.opt("step")
def train_single_forward(metric_func = []):
    def train_single_forward_cb(loop):
        @loop.on_DATA_PROCESS
        def set_xy(loop):
            loop.core.x,loop.core.y = loop.element

        @loop.on_FORWARD
        def forward_pass(loop):
            loop.core.pred = loop.model("__call__",loop.x)[0]

        @loop.on_CALC_LOSS
        def calculate_loss(loop):
            loop.loss["latest"] = loop.loss_func("__call__",loop.pred,loop.y)

        @loop.on_METRICS
        def calcualte_metrics(loop):
            # calculate metrics
            with torch.no_grad():
                for func in metric_func:
                    loop.metrics[func.__name__] = func(loop.pred,loop.y)

            # loop through metrics
            loop.pgbar_data(loop.metrics.cases)

    return train_single_forward_cb

class TrainLoop(LoopStack):
    def __init__(self,data_iter,model=[],opt=[],loss_func=[],loss=[],hp=[],cuda=[],
                 callbacks = [train_callbacks,]):
        self.from_loops(ProgressBar,Tolerate,
                        *events(*TRAIN_EVENTS))
        self.new_setting("model","opt","loss_func","loss","hp","cuda","metric_func","metric")
        self.__init__(data_iter,)
        for cb in callbacks:
            cb(self)

class EvalLoop(LoopStack):
    def __init__(self,data_iter):
        self.from_loops(ProgressBar,Tolerate,
                        *events(*EVAL_EVENTS))
        self.new_setting("model","opt","loss","hp","cuda")
        self.__init__(data_iter,)

        @self.EVAL_FORWARD.downstream
        def torch_eval_wrap(self,func):
            with torch.no_grad():
                func()